#Base HQQ/VLLM model defintion
import torch

from vllm.model_executor.layers.linear import UnquantizedLinearMethod
class HQQLinearMethod(UnquantizedLinearMethod):
	def apply_weights(self, weights, x, bias=None):
		return weights['hqq_module'](x)

from ..base import BaseHQQModel, HQQLinear
class BaseHQQVLLMModel(BaseHQQModel):

	@classmethod
	def quantize_model_single_worker(cls, model, quant_config):
		#Use the same quantization config for all linear layers. Use None to skip quantizing a specfic layer.
		patch_params = dict([(k, quant_config) for k in cls.get_linear_tags()])

		#We replace the nn.Linear layers with HQQLinear
		def _patch_linear(linear_layer, quant_config):
			if(quant_config is None): return linear_layer

			hqq_module = HQQLinear(linear_layer, quant_config, del_orig=False)

			#Clear original params
			del linear_layer.weight
			linear_layer.bias = None #bias is inside hqq_module

			#Set HQQ params 
			linear_layer.linear_weights = {'hqq_module':hqq_module}
			linear_layer.linear_method  = HQQLinearMethod()

			torch.cuda.empty_cache()

			return linear_layer

		cls.patch_model(model, lambda l: l.half().cuda(), _patch_linear, patch_params)

	@classmethod
	def quantize_model(cls, model, quant_config):
		workers = model.llm_engine.workers
		for i in range(len(workers)):
			cls.quantize_model_single_worker(workers[i].model, quant_config=quant_config)

	#Save model architecture
	@classmethod
	def cache_model(cls, model, save_dir):
		model_0 = model.llm_engine.workers[0].model
		model_0.config.save_pretrained(save_dir)

	@classmethod
	def cache_model_single_worker(cls, model_0, save_dir):
		model_0.config.save_pretrained(save_dir)

	@classmethod
	def serialize_weights(cls, model, verbose):
		weights     = {}
		ignore_keys = cls.get_ignore_layers(model)
		for name, module in model.named_modules():
			if(name in ignore_keys): continue
			try:
				state_dict = module.state_dict()
				if(len(state_dict)>0): 
					weights[name] = dict(state_dict)
				else:
					#Quantized linear layers in a VLLM model
					if(hasattr(module, 'linear_weights')):
						if('hqq_module' in module.linear_weights):
							weights[name] = module.linear_weights['hqq_module'].state_dict()

			except Exception as error:
				if(verbose): 
					print('Skipping', name)

		return weights

	@classmethod
	def save_quantized(cls, model, save_dir, verbose=False):
		model_0 = model.llm_engine.workers[0].model

		#Cache model
		cls.cache_model_single_worker(model_0, save_dir)

		#Serialization
		weights = cls.serialize_weights(model_0, verbose=verbose)

		#Save
		cls.save_weights(weights, save_dir)

	#################################################
	@classmethod
	def from_quantized_single_worker(cls, save_dir_or_hub, cache_dir='', device='cuda:0'):
		#Get directory path
		save_dir = cls.try_snapshot_download(save_dir_or_hub, cache_dir)

		#Load model from config
		model = cls.create_model(save_dir)

		#Name the layers
		cls.autoname_modules(model) 

		#Load weights
		try:
			weights = cls.load_weights(save_dir, map_location=device)
		except Exception as error:
			print("Failed to load the weights", error)
			return
		
		#load_state_dict() doesn't work with modules initialized with init_empty_weights(), so we need to do this manually
		@torch.no_grad()
		def _load_module(module, params=None):
			if(module.name not in weights): 
				return module.half().cuda()

			state_dict = weights[module.name]
			if(('W_q' in state_dict) and ('meta' in state_dict)):
				hqq_module = HQQLinear(linear_layer=None, quant_config=None)
				hqq_module.load_state_dict(state_dict)

				#Clear original params
				del module.weight
				module.bias = None #bias is inside hqq_module

				#Set HQQ params 
				module.linear_weights = {'hqq_module':hqq_module}
				module.linear_method  = HQQLinearMethod()

				torch.cuda.empty_cache()

			else:
				for key in state_dict:
					setattr(module, key, torch.nn.Parameter(state_dict[key], requires_grad=False))

			return module 

		#Load modules
		cls.patch_model(model, _load_module, _load_module, dict([(k, None) for k in cls.get_linear_tags()]))
		#Load other weights that are not part of any module
		cls.post_module_load(model, weights) 
		
		return model



#Import VLLM architectures here with the dummy_load trick
import torch, transformers
from typing import Dict
from ..models.vllm.llama import LlamaForCausalLM, LlamaHQQ
from ..models.base import BaseHQQModel
from .base import HQQWrapper

#Set them in the model registry
import vllm 
vllm.model_executor.model_loader._MODEL_REGISTRY['LlamaForCausalLM'] = LlamaForCausalLM

_HQQ_REGISTRY = {}
_HQQ_REGISTRY['LlamaForCausalLM'] = LlamaHQQ

#VLLM requires model input as string, so this is the fallback for a dummy init used to load the model from quantized weights + load the tokenizer
_ARCH_TO_DEFAULT = {}
_ARCH_TO_DEFAULT['LlamaForCausalLM'] = 'meta-llama/Llama-2-7b-chat-hf'

_Parent = vllm.entrypoints.llm.LLM

#Similar to the from vllm import LLM class, with some extra parameters and only loads a dummy model
class HQQLLM(_Parent, HQQWrapper):
	_HQQ_REGISTRY  = _HQQ_REGISTRY
	INIT_CACHE_MEM = 1.0 #Default cache mem in GB. This is extra-memory, but you need it otherwise runtime is slower 
	
	def __init__(self, *args, **kwargs):

		self._validate_params(kwargs)

		if('force_skip' in kwargs):
			force_skip = kwargs['force_skip']
			del kwargs['force_skip']
		else:
			force_skip = False

		#This will init a dummy model
		super().__init__(*args, **kwargs)

		cache_dir  = kwargs['cache_dir'] if ('cache_dir' in kwargs) else ''
		model_id   = kwargs['model'] 
		
		#Here we load the model on CPU so that we can quantize it while avoiding extra GPU cost
		self.arch_key = self.llm_engine.model_config.hf_config.architectures[0]
		model_config  = self.llm_engine.model_config.hf_config
		workers       = self.llm_engine.workers

		if(force_skip==False):
			for i in range(len(workers)):
				workers[i].model.__init__(model_config, dummy_load=False)
				workers[i].model.load_weights(model_name_or_path=model_id, cache_dir=cache_dir)

		self.hqq_quantized = False

	def _validate_params(self, kwargs:Dict):
		if('gpu_memory_utilization' not in kwargs):
			total_gpu_mem = torch.cuda.get_device_properties(0).total_memory/1e9 #GB
			kwargs['gpu_memory_utilization'] = gpu_memory_utilization = HQQLLM.INIT_CACHE_MEM/total_gpu_mem;

	#In case the user wants to use the fp16 model and skip quantization
	def cuda(self):
		self._check_if_already_quantized(self)
		workers = self.llm_engine.workers
		for i in range(len(workers)):
			workers[i].model = workers[i].model.half().cuda(i)
		return self

	def quantize_model(self, quant_config):
		return self.quantize_model_(model=self, quant_config=quant_config)

	def save_quantized(self, save_dir):
		return self.save_quantized_(model=self, save_dir=save_dir) 

	@classmethod
	def _get_arch_key_from_save_dir(cls, save_dir:str):
		config = transformers.AutoConfig.from_pretrained(save_dir)
		return config.architectures[0]

	#This requires custom loading because VLLM requires a str input model to initalize the workers
	@classmethod
	def from_quantized(cls, save_dir_or_hub, cache_dir='', tensor_parallel_size=1):
		assert tensor_parallel_size==1, "Only single GPU is supported."
		#Both local and hub-support
		save_dir = BaseHQQModel.try_snapshot_download(save_dir_or_hub)
		config   = transformers.AutoConfig.from_pretrained(save_dir)
		arch_key = cls._get_arch_key_from_save_dir(save_dir)
		cls._check_arch_support(arch_key)

		#Trick to initialize the tokenizer and a dummy model inside a VLLM instance
		instance = cls(model=_ARCH_TO_DEFAULT[arch_key], force_skip=True, tensor_parallel_size=1) 
		workers  = instance.llm_engine.workers
		for i in range(len(workers)):
			workers[i].model = cls._get_hqq_class(arch_key).from_quantized_single_worker(save_dir_or_hub=save_dir_or_hub, cache_dir=cache_dir, device='cuda:'+str(i))

		cls._make_quantizable(instance, quantized=True)
		return instance

#From https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/llms/vllm.py
from termcolor import colored
try:
	from typing import Dict
	from langchain.llms import VLLM as LangchainVLLMBase
	from langchain_core.pydantic_v1 import Field, root_validator
	class LangchainVLLM(LangchainVLLMBase):
		def set(self, model):
			self.client = model
			return self

		@root_validator()
		def validate_environment(cls, values: Dict) -> Dict:
			return values
except:
	LangchainVLLM = None
	print(colored('Langchain not installed. You can install it via "pip install langchain"','cyan'))






#Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################

import torch
import numpy as np 

#Proximal solver || W - dequantize(quantize(W))||_p^p
@torch.inference_mode()
def optimize_weights_proximal(tensor, scale, zero, min_max, axis=0, device='cuda', opt_params={'lp_norm':0.7, 'beta':1e1, 'kappa':1.01, 'iters':20}, verbose=False):
	lp_norm, beta, kappa, iters = opt_params['lp_norm'], opt_params['beta'], opt_params['kappa'], opt_params['iters']

	dtype  = torch.float16 if (device=='cuda') else torch.float32
	W_f    = tensor.to(dtype).to(device)
	scale  = scale.to(dtype).to(device)
	zero   = zero.to(dtype).to(device)

	if(lp_norm==1):
		shrink_op = lambda x, beta: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - 1./beta) 
	else:
		shrink_op = lambda x, beta,p=lp_norm: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - (1./beta)*torch.pow(torch.abs(x), p-1))

	best_error = 1e4
	for i in range(iters):
		W_q   = torch.round(W_f*scale + zero).clamp(min_max[0], min_max[1])
		W_r   = (W_q - zero)/scale
		W_e   = shrink_op(W_f - W_r, beta)
		zero  = torch.mean(W_q - (W_f - W_e)*scale, axis=axis, keepdim=True)
		beta *= kappa

		current_error = float(torch.abs(W_f - W_r).mean())
		if(verbose): 
			print(i, np.round(current_error, 6))
		if(current_error < best_error):
			best_error = current_error
		else:
			break

	scale = scale.to(tensor.device)
	zero  = zero.to(tensor.device)
	del W_f, W_q, W_r, W_e
	torch.cuda.empty_cache()

	return scale, zero 

#SGD solver  || W - dequantize(quantize(W))||_1 (p=1 only)
def optimize_weights_autograd(tensor, scale, zero, min_max, axis=0, device='cuda', opt_params={'lr':2e-3, 'iters':2500}, verbose=False): 
	W_f             = tensor.to(device) 
	params          = {}
	params['scale'] = torch.nn.Parameter(scale.float().to(device), requires_grad=True)
	params['zero']  = torch.nn.Parameter(zero.float().to(device),  requires_grad=True)
	optimizer       = torch.optim.AdamW([params[k] for k in params], lr=opt_params['lr'], betas=(0.9, 0.99), eps=1e-06, weight_decay=0.)

	def _loss_fct(output, target):
		return torch.mean(torch.abs(target - output)) #L1

	def _fake_quant():
		#Quantize
		W_q = torch.round(W_f*params['scale'] + params['zero']).clamp(min_max[0], min_max[1])
		#Dequantize
		W_r = (W_q - params['zero'])/params['scale'] 
		return W_r

	with torch.no_grad():
		_init_loss = _loss_fct(_fake_quant(), W_f).item()

	def _step():
		optimizer.zero_grad()
		loss    = _loss_fct(_fake_quant(), W_f)
		loss.backward()
		optimizer.step()
		return  np.round(loss.item(), 10)

	for i in range(opt_params['iters']):
		l = _step()
		if(verbose and (i%100)==0): print(i, l)
		
	with torch.no_grad():
		_final_loss = _loss_fct(_fake_quant(), W_f).item()

	if(_final_loss<_init_loss):
		for k in params: params[k] = params[k].data.detach().to(tensor.device)
	else:
		if(verbose): print('optimization failed...')
		params = {'scale':scale, 'zero':zero}

	del W_f	
	torch.cuda.empty_cache()
	return params['scale'], params['zero']



from tqdm import tqdm 
import transformers
from accelerate import init_empty_weights

from ..base import BaseHQQModel
class BaseHQQHFModel(BaseHQQModel):
	#Save model architecture
	@classmethod
	def cache_model(cls, model, save_dir):
		model.config.save_pretrained(save_dir)



from ..base import BaseHQQModel
import os, json, timm

class BaseHQQTimmModel(BaseHQQModel):
	#Save model architecture
	@classmethod
	def cache_model(cls, model, save_dir):
		try:
			os.makedirs(save_dir, exist_ok=True)
		except Exception as error:
			print(error)

		with open(cls.get_config_file(save_dir), "w") as file:
			json.dump(model.default_cfg, file)

	#Create empty model
	@classmethod
	def create_model(cls, save_dir):
		with open(cls.get_config_file(save_dir), "r") as file:
			config = json.load(file)
		model = timm.create_model(config['architecture'] + '.' + config['tag'], pretrained=False)
		return model




from abc import abstractmethod
from typing import Dict
from ..models.base import BaseHQQModel

#Wrapper that makes it easier to add quantization support to different engines (HF, VLLM, etc.)

class HQQWrapper:

	@abstractmethod
	def _get_arch_key_from_save_dir(cls, save_dir:str):
		pass

	@classmethod
	def _get_hqq_class(cls, arg):
		arch = arg if (type(arg)==str) else arg.arch_key
		return cls._HQQ_REGISTRY[arch]

	@classmethod
	def _validate_params(cls, params:Dict):
		pass

	@classmethod
	def _is_quantizable(cls, model):
		return hasattr(model, 'hqq_quantized')

	@classmethod
	def _make_quantizable(cls, model, quantized):
		model.hqq_quantized = quantized

	@classmethod
	def _check_arch_support(cls, arg):
		arch = arg if (type(arg)==str) else arg.arch_key
		assert (arch in cls._HQQ_REGISTRY), "Model architecture " + arch + " not supported yet."

	@classmethod
	def _check_if_already_quantized(cls, model):
		assert (not model.hqq_quantized), "Model already quantized"

	@classmethod
	def _check_if_not_quantized(cls, model):
		assert model.hqq_quantized, "Model not quantized."

	@classmethod
	def _set_quantized(cls, model, quantized):
		model.hqq_quantized = quantized

	#####################################################
	@classmethod
	def quantize_model_(cls, model, quant_config):
		if(cls._is_quantizable(model)==False):
			cls._make_quantizable(model, quantized=False)  
		cls._check_arch_support(model)
		cls._check_if_already_quantized(model)
		cls._get_hqq_class(model).quantize_model(model, quant_config=quant_config)
		cls._set_quantized(model, True)

	@classmethod
	def save_quantized_(cls, model, save_dir):
		cls._check_if_not_quantized(model)
		cls._get_hqq_class(model).save_quantized(model, save_dir=save_dir)

	@classmethod
	def from_quantized(cls, save_dir_or_hub, cache_dir=''):
		#Both local and hub-support
		save_dir = BaseHQQModel.try_snapshot_download(save_dir_or_hub)
		arch_key = cls._get_arch_key_from_save_dir(save_dir)
		cls._check_arch_support(arch_key)

		model = cls._get_hqq_class(arch_key).from_quantized(save_dir, cache_dir)

		cls._make_quantizable(model, quantized=True)
		return model

	@classmethod
	def get_linear_tags(cls, model):
		return cls._get_hqq_class(model).get_linear_tags()




import transformers, json
from typing import Dict

_HQQ_REGISTRY = {}

from ..models.hf.llama import LlamaHQQ
_HQQ_REGISTRY['LlamaForCausalLM'] = LlamaHQQ

from hqq.models.hf.mixtral import MixtralHQQ
_HQQ_REGISTRY['MixtralForCausalLM'] = MixtralHQQ

#Alias 
AutoTokenizer = transformers.AutoTokenizer

#Used to call super() on classmethods
_Parent = transformers.AutoModelForCausalLM

from .base import HQQWrapper

class HQQModelForCausalLM(_Parent, HQQWrapper):
	_HQQ_REGISTRY = _HQQ_REGISTRY

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		cls._make_quantizable(model, quantized=False)

	@classmethod
	def _make_quantizable(cls, model, quantized):
		model.hqq_quantized  = quantized
		model.arch_key       = model.config.architectures[0] 
		model.quantize_model = lambda quant_config: cls.quantize_model_(model=model, quant_config=quant_config)
		model.save_quantized = lambda save_dir: cls.save_quantized_(model=model, save_dir=save_dir)
		model.cuda           = lambda *args, **kwargs: model if(quantized) else model.cuda
		model.to             = lambda *args, **kwargs: model if(quantized) else model.to
		model.float          = lambda *args, **kwargs: model if(quantized) else model.float
		model.half           = lambda *args, **kwargs: model if(quantized) else model.half

	#Force loading the model on CPU and unquantized 
	@classmethod
	def _validate_params(cls, params:Dict):
		for p in ['load_in_4bit', 'load_in_8bit']: #ignore these
			if(p in params):
				params[p] = False 
		params['device_map'] = None

	@classmethod
	def from_pretrained(cls, *args, **kwargs):
		cls._validate_params(kwargs)
		model = super(_Parent, cls).from_pretrained(*args, **kwargs)
		cls._make_quantizable(model, quantized=False)
		return model

	@classmethod
	def _get_arch_key_from_save_dir(cls, save_dir:str):
		config = transformers.AutoConfig.from_pretrained(save_dir)
		return config.architectures[0]




from ..base import *
from .base import BaseHQQTimmModel
from tqdm import tqdm 

#Patch ViT functions
class VitCLIPPatch(BasePatch):
	#These tags are used to specify the parameters of each layer type. For example, if you want to give different quantization parameters to different layers
	@classmethod
	def get_linear_tags(cls):
		return ['attn.qkv', 'attn.proj', 'mlp.fc1', 'mlp.fc2']

	@classmethod
	def freeze_model(cls, model):
		for param in model.parameters():
			param.requires_grad = False

	@classmethod
	def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
		model.patch_embed.proj = patch_fct(model.patch_embed.proj)
		model.patch_embed.norm = patch_fct(model.patch_embed.norm)
		model.norm_pre         = patch_fct(model.norm_pre)
		model.norm             = patch_fct(model.norm)
		model.head             = patch_fct(model.head)
		model.cls_token.data   = patch_fct(model.cls_token.data)
		model.pos_embed.data   = patch_fct(model.pos_embed.data)

		for i in tqdm(range(len(model.blocks)), disable=not verbose):
			model.blocks[i].norm1 = patch_fct(model.blocks[i].norm1)
			model.blocks[i].norm2 = patch_fct(model.blocks[i].norm2)

	@classmethod
	def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
		for i in tqdm(range(len(model.blocks))):
			model.blocks[i].attn.qkv  = patch_fct(model.blocks[i].attn.qkv,  patch_params['attn.qkv']) 
			model.blocks[i].attn.proj = patch_fct(model.blocks[i].attn.proj, patch_params['attn.proj'])
			model.blocks[i].mlp.fc1   = patch_fct(model.blocks[i].mlp.fc1,   patch_params['mlp.fc1']) 
			model.blocks[i].mlp.fc2   = patch_fct(model.blocks[i].mlp.fc2,   patch_params['mlp.fc2']) 


class ViTCLIPHQQ(VitCLIPPatch, BaseHQQTimmModel):
	#layers to ignore when saving the weights
	@classmethod
	def get_ignore_layers(cls, model):
		return ['', 'model', 'model.blocks'] + ['model.blocks.' + str(i) for i in range(len(model.blocks))]

	#since cls_token and pos_embed are trainable parameters but are not part of any module, we need to add them manually 
	#for saving
	@classmethod
	def serialize_weights(cls, model, verbose):
		weights = super().serialize_weights(model, verbose) 
		weights['cls_token'] = model.cls_token.data
		weights['pos_embed'] = model.pos_embed.data
		return weights

	#and loading
	@classmethod
	def post_module_load(cls, model, weights):
		super().post_module_load(model, weights) 
		model.cls_token.data = weights['cls_token']
		model.pos_embed.data = weights['pos_embed']
		



#Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch 
import gc, os, copy
from tqdm import tqdm 
from abc import abstractmethod

from huggingface_hub import snapshot_download
from ..core.quantize import HQQLinear

#Cleanup GPU vram
def cleanup():
	torch.cuda.empty_cache()
	gc.collect()

#Make sure file paths end with '/'
def fix_path(path):
	if(len(path)==0): return path
	return path + '/' if (path[-1]!='/') else path 

#Base patching class. Patching defines how nn.Linear and other layers are replaced via a patching function. 
class BasePatch():
	#Override these OR override the main patch_model() function
	############################################
	#This method iterates through layers of the model that are NOT nn.Linear and processes them via new_nodule = patch_fct(module, params)
	@classmethod
	def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
		pass

	#This method iterates through layers of the model that are nn.Linear and processes them via new_nodule = patch_fct(module, params)
	@classmethod
	def patch_linearlayers(cls, base_model, patch_fct, patch_params, verbose=True):
		pass 
	############################################
	#These tags are used to specfiy parameters of the patching in  patch_linearlayers()
	@classmethod
	def get_linear_tags(cls):
		return []
	
	#Autmatically name modules. This is very important to save/load the weights 
	@classmethod
	def autoname_modules(cls, model):
		for name, module in model.named_modules():
			module.name = name

	#Freeze all layers
	@classmethod
	def freeze_model(cls, model):
		for param in model.parameters():
			param.requires_grad = False
		try:
			for param in model.model.parameters():
				param.requires_grad = False
		except:
			pass

	#Main patching function
	@classmethod
	def patch_model(cls, model, patch_nonlinear_fct, patch_linear_fct, patch_params, verbose=True):
		model.eval()
		cls.freeze_model(model)
		cls.patch_nonlinearlayers(model, patch_nonlinear_fct, verbose=verbose)
		cls.patch_linearlayers(model, patch_linear_fct, patch_params, verbose=verbose)
		cls.autoname_modules(model)
		cleanup()


class BaseHQQModel:
	#Override these
	############################################
	#This method creates and empty model based on the specfied architecture
	@abstractmethod
	def create_model(self):
		pass

	#This method saves the model architecture only without inculding the weights (for example to a config.json)
	@abstractmethod	
	def cache_model(cls, model, save_dir):
		pass
	############################################

	@classmethod
	def get_config_file(cls, save_dir):
		return fix_path(save_dir) + 'config.json'

	@classmethod
	def get_weight_file(cls, save_dir):
		return fix_path(save_dir) + 'qmodel.pt'    

	@classmethod
	def get_ignore_layers(cls, model):
		return []

	#Save weights to disk
	@classmethod
	def save_weights(cls, weights, save_dir):
		torch.save(weights, cls.get_weight_file(save_dir))

	#Load weights from disk
	@classmethod
	def load_weights(cls, save_dir, map_location=None):
		return torch.load(cls.get_weight_file(save_dir), map_location=map_location)

	#Main function to quantize a model. Basically goes through the linear layers specfied in the patching function and replaces them with HQQLinear
	@classmethod
	def quantize_model(cls, model, quant_config):
		#Use the same quantization config for all linear layers. Use None to skip quantizing a specfic layer.
		if(True in [(key in cls.get_linear_tags()) for key in quant_config.keys()]): 
			#If the user doesn't specify a key from get_linear_tags, the layer is not quantized via (key, None)
			patch_params = dict([(key, None) for key in cls.get_linear_tags()])
			patch_params.update(quant_config)
		else:
			#Same quant_config for all layers
			patch_params = dict([(k, quant_config) for k in cls.get_linear_tags()])

		#We replace the nn.Linear layers with HQQLinear
		def _patch_linear(linear_layer, quant_config):
			return HQQLinear(linear_layer, quant_config) if (quant_config is not None) else linear_layer.half().cuda()

		cls.patch_model(model, lambda l: l.half().cuda(), _patch_linear, patch_params)

	#Prepares model weights by iterating through modules. It might some parameters that are NOT modules like model.param1
	@classmethod
	def serialize_weights(cls, model, verbose):
		weights     = {}
		ignore_keys = cls.get_ignore_layers(model)
		for name, module in model.named_modules():
			if(name in ignore_keys): continue
			try:
				state_dict = module.state_dict()
				if(len(state_dict)>0): 
					weights[name] = dict(state_dict)
			except Exception as error:
				if(verbose): 
					print('Skipping', name)

		return weights

	#Main function to save a quantized model
	@classmethod
	def save_quantized(cls, model, save_dir, verbose=False):
		#Save config
		cls.cache_model(model, save_dir)

		#Serialization
		weights = cls.serialize_weights(model, verbose=verbose)

		#Save
		cls.save_weights(weights, save_dir)

	@classmethod
	def try_snapshot_download(cls, save_dir_or_hub, cache_dir=''):
		save_dir = fix_path(cache_dir) + save_dir_or_hub

		if(os.path.exists(save_dir)==False):
			save_dir = snapshot_download(repo_id=save_dir_or_hub, cache_dir=cache_dir)
			save_dir = fix_path(save_dir)

		#Check 
		if(os.path.exists(cls.get_weight_file(save_dir))==False):
			raise Exception('Weight file missing. Check your cache directory.')
		if(os.path.exists(cls.get_config_file(save_dir))==False):
			raise Exception('Config file missing. Check your cache directory.')

		return save_dir


	#This method is specfically designed in case we need to load some weights that are not part of any module
	@classmethod
	def post_module_load(cls, model, weights):
		pass

	#Main function to load an HQQ quantized model from either HF hub or locally
	@classmethod
	def from_quantized(cls, save_dir_or_hub, cache_dir=''):
		#Get directory path
		save_dir = cls.try_snapshot_download(save_dir_or_hub, cache_dir)

		#Load model from config
		model = cls.create_model(save_dir)

		#Name the layers
		cls.autoname_modules(model) 

		#Load weights
		try:
			weights = cls.load_weights(save_dir)
		except Exception as error:
			print("Failed to load the weights", error)
			return
		
		#load_state_dict() doesn't work with modules initialized with init_empty_weights(), so we need to do this manually
		@torch.no_grad()
		def _load_module(module, params=None):
			if(module.name not in weights): 
				return module.half().cuda()

			state_dict = weights[module.name]
			if(('W_q' in state_dict) and ('meta' in state_dict)):
				module = HQQLinear(linear_layer=None, quant_config=None)
				module.load_state_dict(state_dict)
			else:
				for key in state_dict:
					setattr(module, key, torch.nn.Parameter(state_dict[key], requires_grad=False))

			return module 

		#Load modules
		cls.patch_model(model, _load_module, _load_module, dict([(k, None) for k in cls.get_linear_tags()]))
		#Load other weights that are not part of any module
		cls.post_module_load(model, weights) 
		
		return model






#Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
import numpy as np 

from .utils    import *
from .optimize import *
from .bitpack  import BitPack 
from tqdm import tqdm
from termcolor import colored

#Main HQQ Quantizer 
class Quantizer:
	SUPPORTED_BITS   = [8, 4, 3, 2]
	optimize_weights = optimize_weights_proximal

	bit_to_packing   = {8:'8bit_u8', 4:'4bit_u8', 3:'3bit_32', 2:'2bit_u8'}

	pack  =  {'8bit_u8':BitPack.pack_8bit_u8,
			  '4bit_u8':BitPack.pack_4bit_u8,
			  '3bit_32':BitPack.pack_3bit_32,
			  '2bit_u8':BitPack.pack_2bit_u8}

	unpack = {'8bit_u8':BitPack.unpack_8bit_u8,
			  '4bit_u8':BitPack.unpack_4bit_u8,
			  '3bit_32':BitPack.unpack_3bit_32,
			  '2bit_u8':BitPack.unpack_2bit_u8}

	@classmethod
	def quantize(cls, tensor, nbits=4, channel_wise=True, group_size=64, optimize=False, round_zero=False, axis=0):
		assert nbits in Quantizer.SUPPORTED_BITS, "nbits=" + str(nbits) + " not supported."
		assert axis in [0, 1], "axis should be either 0 or 1"
		if(group_size is not None):
			assert is_divisible(tensor.numel(), group_size), "group_size should be divisble by the total tensor dimensions. shape: " + str(tensor.shape) + ", group_size: " + str(group_size)

		W     = tensor.float() 
		shape = W.shape

		#Reshape for grouping
		if((group_size is not None) and channel_wise):
			W = W.reshape([-1, group_size]) if (axis==1) else W.reshape([group_size, -1])
		
		#Get min/max values
		if(channel_wise==False):
			_min, _max = W.min(), W.max()
			optimize   = False
		else:
			_min  = W.min(axis=axis, keepdim=True)[0]
			_max  = W.max(axis=axis, keepdim=True)[0]

		max_v   = 2**nbits - 1 
		min_v   = 0
		min_max = [min_v, max_v]

		#Note: here we work with the inverse of the scale to avoid division and quantize instead via W*scale + zero, the scale is inverted later on.
		scale   = (max_v/(_max - _min)).clamp(max=2e4) #clamp to avoid half-precision problems
		zero    = -_min*scale 

		#Round zero as in: https://github.com/casper-hansen/AutoAWQ/blob/main/awq/quantize/quantizer.py#L42C9-L42C14
		if(round_zero): zero = torch.round(zero)
		
		#Fine-tune weights
		if(optimize): scale, zero = Quantizer.optimize_weights(tensor=W, scale=scale, zero=zero, min_max=min_max, axis=axis)

		#Quantize
		W_q  = torch.round(W*scale + zero).clamp(min_max[0], min_max[1])

		#Store meta-data (we invert the scale for dequantization)
		meta = {'nbits':nbits, 'group_size':group_size, 'shape':shape, 'scale':1./scale, 'zero':zero, 'axis':axis, 'packing':Quantizer.bit_to_packing[nbits]}

		#Pack bits
		W_q  = Quantizer.pack[meta['packing']](W_q)

		#cleanup
		del W, _min, _max 
		torch.cuda.empty_cache()

		return W_q, meta

	#Main dequantization: bit_unpacking > (W_q - z)*s > reshape
	@classmethod
	def dequantize(cls, W_q, meta):
		W_q_p = Quantizer.unpack[meta['packing']](W_q).half()
		if((meta['group_size'] is not None) and (meta['nbits']==3)):
			W_q_p = W_q_p[:meta['group_size']] if (meta['axis']==0) else W_q_p[:,:meta['group_size']]
		W_r = ((W_q_p - meta['zero'])*meta['scale']).reshape(meta['shape']) 
		del W_q_p
		return W_r

	@classmethod
	def to_inplace(cls, W_q, meta, device):
		W_q = W_q.to(device).contiguous() 
		for key in meta:
			if(type(meta[key])==torch.Tensor):
				meta[key] = (meta[key].half() if meta[key].dtype==torch.float32 else meta[key]).to(device).contiguous() 
		return W_q, meta

	@classmethod
	def to_ooplace(cls, W_q, meta, device):
		W_q_c  = W_q.to(device).contiguous() 
		meta_c = {}
		for key in meta:
			if(type(meta[key])==torch.Tensor):
				meta_c[key] = (meta[key].half() if meta[key].dtype==torch.float32 else meta[key]).to(device).contiguous() 
			else:
				meta_c[key] = meta[key]
		return W_q_c, meta_c

	@classmethod
	def cuda(cls, W_q, meta, device_n=0):
		return Quantizer.to_inplace(W_q, meta, device='cuda:' + str(device_n))

	@classmethod
	def cpu(cls, W_q, meta):
		return Quantizer.to_ooplace(W_q, meta, device='cpu')


#Main linear layer 
try:
	import hqq_aten
except:
	print(colored('hqq_aten package not installed. HQQBackend.ATEN backend will not work unless you install the hqq_aten lib in hqq/kernels.', 'cyan'))
	hqq_aten = None

from enum import Enum
class HQQBackend(Enum):
	#Name of the forward functions
	PYTORCH         = "forward_pytorch" 
	PYTORCH_COMPILE = "forward_pytorch_compile"
	ATEN            = "forward_aten"

#Main linear layer 
class HQQLinear(torch.nn.Module):
	backend = HQQBackend.PYTORCH #Default

	def __init__(self, linear_layer, quant_config, del_orig=True):
		super().__init__()
		self.ready        = False
		self.in_gpu       = False
		self.quant_config = quant_config
		self.set_backend(HQQLinear.backend) #Default backend
		if(linear_layer is not None):
			self.bias = None if (linear_layer.bias==None) else linear_layer.bias.half().cuda()
			self.quantize(linear_layer.weight.data, **quant_config)
		if(del_orig): del linear_layer
		torch.cuda.empty_cache()
		
	def cuda(self, device_n=0):
		if(self.in_gpu): return 
		self.W_q, self.meta = Quantizer.cuda(self.W_q, self.meta, device_n)
		if(self.meta['quant_scale']):
			self.meta['scale_q'] , self.meta['meta_scale'] = Quantizer.cuda(self.meta['scale_q'], self.meta['meta_scale'], device_n)
		if(self.meta['quant_zero']):
			self.meta['zero_q'] , self.meta['meta_zero']   = Quantizer.cuda(self.meta['zero_q'], self.meta['meta_zero'], device_n)

		if(self.bias is not None):
			self.bias = self.bias.half().cuda(device_n)

		self.in_gpu = True

	def to(self, device):
		pass

	def half(self):
		return self 

	def state_dict(self):
		return {'W_q':self.W_q, 'meta':self.meta, 'bias':self.bias}

	def load_state_dict(self, state_dict):
		self.W_q    = state_dict['W_q']
		self.meta   = state_dict['meta']
		self.bias   = state_dict['bias'] if ('bias' in state_dict) else None
		self.in_gpu = self.W_q.device.type == 'cuda'
		if(self.in_gpu==False): self.cuda()
		self.ready  = True

	def quantize(self, W, weight_quant_params, scale_quant_params, zero_quant_params):
		quant_scale = scale_quant_params is not None
		quant_zero  = zero_quant_params  is not None
		
		#Quantize
		W_q , meta = Quantizer.quantize(W, **weight_quant_params) 
		meta.update({'quant_scale':quant_scale, 'quant_zero':quant_zero})
		if(meta['quant_scale']):
			meta['scale_q'] , meta['meta_scale'] = Quantizer.quantize(meta['scale'], **scale_quant_params); del meta['scale']
		if(meta['quant_zero']):
			meta['zero_q'], meta['meta_zero']    = Quantizer.quantize(meta['zero'],  **zero_quant_params);  del meta['zero']

		self.W_q   = W_q
		self.meta  = meta 
		self.cuda()
		self.ready = True

	@torch.inference_mode()
	def dequantize(self):
		assert self.ready, "model was not quantized"
		W_q, meta = self.W_q, self.meta
		del_keys = []
		if(meta['quant_scale']):
			meta['scale'] = Quantizer.dequantize(meta['scale_q'], meta['meta_scale']); del_keys.append('scale')
		if(meta['quant_zero']):
			meta['zero']  = Quantizer.dequantize(meta['zero_q'],  meta['meta_zero']);  del_keys.append('zero')
		W_est = Quantizer.dequantize(W_q, meta)
		#Cleanup
		for key in del_keys: del meta[key]
		return W_est

	@classmethod
	def set_backend(cls, backend: HQQBackend):
		HQQLinear.backend = backend
		cls.forward       = getattr(cls, HQQLinear.backend.value)

	@torch.no_grad()
	def forward_pytorch(self, x):
		W_est = self.dequantize()
		out   = torch.matmul(x, W_est.t())
		if(self.bias!=None): out += self.bias
		del W_est
		return out

	##############################################
	#Experimental
	#############################################
	@torch.no_grad()
	@torch.compile()
	def forward_pytorch_compile(self, x):
		W_est = self.dequantize()
		out   = torch.matmul(x, W_est.t())
		if(self.bias!=None): out += self.bias
		del W_est
		return out

	@torch.no_grad()
	def forward_aten(self, x):
		empt = torch.empty([0])
		W_q  = self.W_q
		meta = self.meta
		bias = self.bias

		W_q, W_s, W_z              = W_q,  empt if (meta['quant_scale']) else meta['scale'], empt if (meta['quant_zero']) else meta['zero']
		W_shape,  W_group_size     = meta['shape'], meta['group_size']
		W_nbits, W_axis, W_packing = meta['nbits'], meta['axis'], meta['packing']

		if(meta['quant_scale']):
			S_q, S_s, S_z              = meta['scale_q'],             meta['meta_scale']['scale'], meta['meta_scale']['zero']
			S_shape, S_group_size      = meta['meta_scale']['shape'], meta['meta_scale']['group_size'] 
			S_nbits, S_axis, S_packing = meta['meta_scale']['nbits'], meta['meta_scale']['axis'],  meta['meta_scale']['packing']
		else:
			S_q, S_s, S_z              = empt, empt, empt
			S_shape, S_group_size      = meta['shape'], -1
			S_nbits, S_axis, S_packing = -1, 0, ""

		if(meta['quant_zero']):
			Z_q, Z_s, Z_z              = meta['zero_q'],             meta['meta_zero']['scale'], meta['meta_zero']['zero']
			Z_shape, Z_group_size      = meta['meta_zero']['shape'], meta['meta_zero']['group_size']
			Z_nbits, Z_axis, Z_packing = meta['meta_zero']['nbits'], meta['meta_zero']['axis'],  meta['meta_zero']['packing']
		else:
			S_q, S_s, S_z              = empt, empt, empt
			S_shape, S_group_size      = meta['shape'], -1
			S_nbits, S_axis, S_packing = -1, 0, ""


		S_group_size = 0 if (S_group_size==None) else S_group_size
		Z_group_size = 0 if (Z_group_size==None) else Z_group_size

		args = [x, bias if (bias is not None) else empt,
				W_q, W_s, W_z, W_shape, W_group_size, W_nbits, W_axis, W_packing,
				S_q, S_s, S_z, S_shape, S_group_size, S_nbits, S_axis, S_packing,
				Z_q, Z_s, Z_z, Z_shape, Z_group_size, Z_nbits, Z_axis, Z_packing]

		return hqq_aten.forward_with_quant(*args)


def hqq_base_quant_config(nbits=4, group_size=64, quant_zero=True, quant_scale=False):
	assert nbits in Quantizer.SUPPORTED_BITS, "nbits value not supported. Check Quantizer.SUPPORTED_BITS."
	assert is_divisible(group_size, 8), "Invalid group_size param: the value should be a multiple of 8." 
	weight_quant_params = {'nbits':nbits,'channel_wise':True,  'group_size':group_size, 'optimize':True, 'round_zero':True if nbits==4 else False} 
	scale_quant_params  = {'nbits':8,    'channel_wise':True,  'group_size':128,        'optimize':False} if (quant_scale) else None
	zero_quant_params   = {'nbits':8,    'channel_wise':False, 'group_size':None,       'optimize':False} if (quant_zero)  else None
	return {'weight_quant_params':weight_quant_params, 'scale_quant_params':scale_quant_params, 'zero_quant_params':zero_quant_params}

#Alias: follow similar Auto-GPTQ naming
BaseQuantizeConfig = hqq_base_quant_config


import timm, json
from typing import Dict
from ..models.base import BaseHQQModel
from ..models.timm.vit_clip import ViTCLIPHQQ

_HQQ_REGISTRY = {}
_HQQ_REGISTRY['vit_huge_patch14_clip_336']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_huge_patch14_clip_224']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_large_patch14_clip_224'] = ViTCLIPHQQ
_HQQ_REGISTRY['vit_large_patch14_clip_336'] = ViTCLIPHQQ
_HQQ_REGISTRY['vit_base_patch16_clip_384']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_base_patch32_clip_448']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_base_patch32_clip_384']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_base_patch16_clip_224']  = ViTCLIPHQQ
_HQQ_REGISTRY['vit_base_patch32_clip_224']  = ViTCLIPHQQ

from .base import HQQWrapper
class HQQtimm(HQQWrapper):
	_HQQ_REGISTRY = _HQQ_REGISTRY

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		cls._make_quantizable(model, quantized=False)

	@classmethod
	def _make_quantizable(cls, model, quantized):
		model.hqq_quantized  = quantized
		model.arch_key       = model.default_cfg['architecture']
		model.quantize_model = lambda quant_config: cls.quantize_model_(model=model, quant_config=quant_config)
		model.save_quantized = lambda save_dir: cls.save_quantized_(model=model, save_dir=save_dir)
		model.cuda           = lambda *args, **kwargs: model if(quantized) else model.cuda
		model.to             = lambda *args, **kwargs: model if(quantized) else model.to
		model.float          = lambda *args, **kwargs: model if(quantized) else model.float
		model.half           = lambda *args, **kwargs: model if(quantized) else model.half

	@classmethod
	def _validate_params(cls, params:Dict):
		pass

	@classmethod
	def create_model(cls, *args, **kwargs):
		cls._validate_params(kwargs)
		model = timm.create_model(*args, **kwargs)
		cls._make_quantizable(model, quantized=False)
		return model

	@classmethod
	def _get_arch_key_from_save_dir(cls, save_dir:str):
		with open(BaseHQQModel.get_config_file(save_dir), "r") as file:
			config = json.load(file)
		return config['architecture']






import torch
import numpy as np 

import gc
def cleanup():
	torch.cuda.empty_cache()
	gc.collect()
	
def is_divisible(val1, val2):
	return int(val2*np.ceil(val1/val2))==val1

def make_multiple(val, multiple):
	return int(multiple*np.ceil(val/float(multiple)))

def zero_pad_row(tensor, num_rows, dtype=None):
	out = torch.zeros([num_rows, tensor.shape[1]], device=tensor.device, dtype=tensor.dtype if (dtype is None) else dtype)
	out[:len(tensor)] = tensor
	return W_q



# coding=utf-8
# Adapted from
# https://github.com/huggingface/transformers/blob/v4.28.0/src/transformers/models/llama/modeling_llama.py
# Copyright 2023 The vLLM team.
# Copyright 2022 EleutherAI and the HuggingFace Inc. team. All rights reserved.
#
# This code is based on EleutherAI's GPT-NeoX library and the GPT-NeoX
# and OPT implementations in this library. It has been modified from its
# original forms to accommodate minor architectural differences compared
# to GPT-NeoX and OPT used by the Meta AI team that trained the model.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Inference-only LLaMA model compatible with HuggingFace weights."""
from typing import Any, Dict, List, Optional, Tuple

import torch
from torch import nn
from transformers import LlamaConfig

from vllm.model_executor.input_metadata import InputMetadata
from vllm.model_executor.layers.activation import SiluAndMul
from vllm.model_executor.layers.attention import PagedAttentionWithRoPE
from vllm.model_executor.layers.layernorm import RMSNorm
from vllm.model_executor.layers.linear import (LinearMethodBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               RowParallelLinear)
from vllm.model_executor.layers.sampler import Sampler
from vllm.model_executor.layers.vocab_parallel_embedding import (
    VocabParallelEmbedding, ParallelLMHead)
from vllm.model_executor.parallel_utils.parallel_state import (
    get_tensor_model_parallel_world_size)
from vllm.model_executor.weight_utils import (default_weight_loader,
                                              hf_model_weights_iterator)
from vllm.sequence import SamplerOutput

KVCache = Tuple[torch.Tensor, torch.Tensor]

import gc

class LlamaMLP(nn.Module):

    def __init__(self, hidden_size: int, intermediate_size: int, hidden_act: str, linear_method: Optional[LinearMethodBase] = None,) -> None:
        super().__init__()
        
        ##############################################################################################
        self.gate_up_proj = MergedColumnParallelLinear(hidden_size, [intermediate_size] * 2, bias=False, linear_method=linear_method)
        self.down_proj    = RowParallelLinear(intermediate_size, hidden_size, bias=False, linear_method=linear_method)

        self.gate_up_proj = self.gate_up_proj.cpu()
        self.down_proj    = self.down_proj.cpu()
        torch.cuda.empty_cache()
        ##############################################################################################

        if hidden_act != "silu":
            raise ValueError(f"Unsupported activation: {hidden_act}. ""Only silu is supported for now.")
        self.act_fn = SiluAndMul()

    def forward(self, x):
        gate_up, _ = self.gate_up_proj(x)
        x          = self.act_fn(gate_up)
        x, _       = self.down_proj(x)
        return x


class LlamaAttention(nn.Module):
    def __init__(self, hidden_size: int, num_heads: int, num_kv_heads: int, rope_theta: float = 10000, rope_scaling: Optional[Dict[str, Any]] = None,
        max_position_embeddings: int = 8192, linear_method: Optional[LinearMethodBase] = None,) -> None:
        super().__init__()
        self.hidden_size = hidden_size
        tp_size = get_tensor_model_parallel_world_size()
        self.total_num_heads = num_heads
        assert self.total_num_heads % tp_size == 0
        self.num_heads = self.total_num_heads // tp_size
        self.total_num_kv_heads = num_kv_heads
        if self.total_num_kv_heads >= tp_size:
            # Number of KV heads is greater than TP size, so we partition
            # the KV heads across multiple tensor parallel GPUs.
            assert self.total_num_kv_heads % tp_size == 0
        else:
            # Number of KV heads is less than TP size, so we replicate
            # the KV heads across multiple tensor parallel GPUs.
            assert tp_size % self.total_num_kv_heads == 0
        self.num_kv_heads = max(1, self.total_num_kv_heads // tp_size)
        self.head_dim = hidden_size // self.total_num_heads
        self.q_size = self.num_heads * self.head_dim
        self.kv_size = self.num_kv_heads * self.head_dim
        self.scaling = self.head_dim**-0.5
        self.rope_theta = rope_theta
        self.max_position_embeddings = max_position_embeddings

        ##############################################################################################
        self.qkv_proj = QKVParallelLinear(hidden_size, self.head_dim, self.total_num_heads, self.total_num_kv_heads, bias=False, linear_method=linear_method)

        self.o_proj   = RowParallelLinear(self.total_num_heads * self.head_dim, hidden_size, bias=False, linear_method=linear_method)

        self.attn     = PagedAttentionWithRoPE(self.num_heads, self.head_dim, self.scaling, base=self.rope_theta,
                                           max_position=self.max_position_embeddings, rotary_dim=self.head_dim, 
                                           num_kv_heads=self.num_kv_heads, rope_scaling=rope_scaling)

        self.qkv_proj = self.qkv_proj.cpu()
        self.o_proj   = self.o_proj.cpu()
        torch.cuda.empty_cache()
        ##############################################################################################

    def forward(self, positions: torch.Tensor, hidden_states: torch.Tensor, kv_cache: KVCache, input_metadata: InputMetadata, cache_event: Optional[torch.cuda.Event]) -> torch.Tensor:
        qkv, _           = self.qkv_proj(hidden_states)
        q, k, v          = qkv.split([self.q_size, self.kv_size, self.kv_size], dim=-1)
        k_cache, v_cache = kv_cache
        attn_output      = self.attn(positions, q, k, v, k_cache, v_cache, input_metadata, cache_event)
        output, _        = self.o_proj(attn_output)
        return output


class LlamaDecoderLayer(nn.Module):
    def __init__(self, config: LlamaConfig, linear_method: Optional[LinearMethodBase] = None, ) -> None:
        super().__init__()
        self.hidden_size = config.hidden_size
        rope_theta = getattr(config, "rope_theta", 10000)
        rope_scaling = getattr(config, "rope_scaling", None)
        max_position_embeddings = getattr(config, "max_position_embeddings", 8192)

        self.self_attn = LlamaAttention(
            hidden_size=self.hidden_size,
            num_heads=config.num_attention_heads,
            num_kv_heads=config.num_key_value_heads,
            rope_theta=rope_theta,
            rope_scaling=rope_scaling,
            max_position_embeddings=max_position_embeddings,
            linear_method=linear_method,
        )
        self.mlp = LlamaMLP(
            hidden_size=self.hidden_size,
            intermediate_size=config.intermediate_size,
            hidden_act=config.hidden_act,
            linear_method=linear_method,
        )
        self.input_layernorm         = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def forward(self,positions: torch.Tensor,hidden_states: torch.Tensor, kv_cache: KVCache, input_metadata: InputMetadata, 
                cache_event: Optional[torch.cuda.Event], residual: Optional[torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        # Self Attention
        if residual is None:
            residual = hidden_states
            hidden_states = self.input_layernorm(hidden_states)
        else:
            hidden_states, residual = self.input_layernorm(hidden_states, residual)
        
        hidden_states = self.self_attn(positions=positions, hidden_states=hidden_states, kv_cache=kv_cache, input_metadata=input_metadata, cache_event=cache_event)

        # Fully Connected
        hidden_states, residual = self.post_attention_layernorm(hidden_states, residual)
        hidden_states           = self.mlp(hidden_states)
        return hidden_states, residual


class LlamaModel(nn.Module):

    def __init__(self, config: LlamaConfig, linear_method: Optional[LinearMethodBase] = None,) -> None:
        super().__init__()
        self.config = config
        self.padding_idx  = config.pad_token_id
        self.vocab_size   = config.vocab_size
        self.embed_tokens = VocabParallelEmbedding(config.vocab_size,config.hidden_size,)
        self.layers       = nn.ModuleList([LlamaDecoderLayer(config, linear_method) for _ in range(config.num_hidden_layers)])
        self.norm         = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def forward(
        self,
        input_ids: torch.Tensor,
        positions: torch.Tensor,
        kv_caches: List[KVCache],
        input_metadata: InputMetadata,
        cache_events: Optional[List[torch.cuda.Event]],
    ) -> torch.Tensor:
        hidden_states = self.embed_tokens(input_ids)
        residual      = None
        for i in range(len(self.layers)):
            cache_event             = None if cache_events is None else cache_events[i]
            layer                   = self.layers[i]
            hidden_states, residual = layer(positions, hidden_states, kv_caches[i], input_metadata, cache_event, residual)

        hidden_states, _ = self.norm(hidden_states, residual)
        return hidden_states

class LlamaForCausalLM(nn.Module):

    def __init__(self, config: LlamaConfig, linear_method: Optional[LinearMethodBase] = None, dummy_load: bool = True) -> None:
        super().__init__()

        self.config        = config
        self.linear_method = linear_method

        #Dummy loading Added
        self.dummy_load    = dummy_load
        if(self.dummy_load): return

        self.model         = LlamaModel(config, linear_method)
        self.lm_head       = ParallelLMHead(config.vocab_size, config.hidden_size)
        self.sampler       = Sampler(config.vocab_size)

    def forward(self, input_ids: torch.Tensor, positions: torch.Tensor, kv_caches: List[KVCache], input_metadata: InputMetadata, cache_events: Optional[List[torch.cuda.Event]]) -> SamplerOutput:
        if(self.dummy_load): return torch.empty([0]) #Added

        hidden_states = self.model(input_ids, positions, kv_caches, input_metadata, cache_events)
        next_tokens   = self.sampler(self.lm_head.weight, hidden_states, input_metadata)
        return next_tokens
    
    def load_weights(self, model_name_or_path: str, cache_dir: Optional[str] = None, load_format: str = "auto", revision: Optional[str] = None):
        if(self.dummy_load): return #Added

        stacked_params_mapping = [
            # (param_name, shard_name, shard_id)
            ("qkv_proj", "q_proj", "q"),
            ("qkv_proj", "k_proj", "k"),
            ("qkv_proj", "v_proj", "v"),
            ("gate_up_proj", "gate_proj", 0),
            ("gate_up_proj", "up_proj", 1),
        ]
        params_dict = dict(self.named_parameters())
        for name, loaded_weight in hf_model_weights_iterator(model_name_or_path, cache_dir, load_format, revision):
            if "rotary_emb.inv_freq" in name:
                continue
            for (param_name, weight_name, shard_id) in stacked_params_mapping:
                if weight_name not in name:
                    continue
                param         = params_dict[name.replace(weight_name, param_name)]
                weight_loader = param.weight_loader
                weight_loader(param, loaded_weight, shard_id)
                break
            else:
                param         = params_dict[name]
                weight_loader = getattr(param, "weight_loader", default_weight_loader)
                weight_loader(param, loaded_weight)


#############################################################################################################################################
#############################################################################################################################################
from tqdm import tqdm
import transformers

from ..base import BasePatch 
class LLamaPatch(BasePatch):
    @classmethod
    def get_linear_tags(cls):
        return ['self_attn.qkv_proj', 'self_attn.o_proj', 'mlp.gate_up_proj', 'mlp.down_proj']

    @classmethod
    def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
        base_model              = model.model
        model.sampler           = patch_fct(model.sampler)
        model.lm_head           = patch_fct(model.lm_head)
        base_model.embed_tokens = patch_fct(base_model.embed_tokens)
        base_model.norm         = patch_fct(base_model.norm)

        layers = base_model.layers
        for i in tqdm(range(len(base_model.layers)), disable=not verbose):
            #rotary embed 
            layers[i].self_attn.attn.rotary_emb.cos_sin_cache = torch.nn.Parameter(layers[i].self_attn.attn.rotary_emb.cos_sin_cache, requires_grad=False)
            layers[i].self_attn.attn.rotary_emb = patch_fct(layers[i].self_attn.attn.rotary_emb) 
            layers[i].mlp.act_fn                = patch_fct(layers[i].mlp.act_fn)
            layers[i].input_layernorm           = patch_fct(layers[i].input_layernorm)
            layers[i].post_attention_layernorm  = patch_fct(layers[i].post_attention_layernorm)

    @classmethod
    def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
        base_model = model.model
        layers     = base_model.layers 
        for i in tqdm(range(len(layers)), disable=not verbose):
            layers[i].self_attn.qkv_proj = patch_fct(layers[i].self_attn.qkv_proj, patch_params['self_attn.qkv_proj'])
            layers[i].self_attn.o_proj   = patch_fct(layers[i].self_attn.o_proj,   patch_params['self_attn.o_proj'])
            layers[i].mlp.gate_up_proj   = patch_fct(layers[i].mlp.gate_up_proj,   patch_params['mlp.gate_up_proj'])
            layers[i].mlp.down_proj      = patch_fct(layers[i].mlp.down_proj,      patch_params['mlp.down_proj'])


from .base import BaseHQQVLLMModel
#from ..models.hf.base import init_empty_weights

class LlamaHQQ(LLamaPatch, BaseHQQVLLMModel):
    #layers to ignore when saving the weights
    @classmethod
    def get_ignore_layers(cls, model):
        _tags  = ['', 'model', 'model.layers']
        _tags += ['model.layers.' + str(i) for i in range(len(model.model.layers))]
        _tags += ['model.layers.' + str(i) + '.self_attn' for i in range(len(model.model.layers))] 
        _tags += ['model.layers.' + str(i) + '.self_attn.attn' for i in range(len(model.model.layers))] 
        _tags += ['model.layers.' + str(i) + '.mlp' for i in range(len(model.model.layers))] 
        return _tags

    #Create empty model
    @classmethod
    def create_model(cls, save_dir):
        config = transformers.AutoConfig.from_pretrained(cls.get_config_file(save_dir))
       #with init_empty_weights(): VLLM models fail with init_empty_weights(), so we load them on the CPU instead
        model = LlamaForCausalLM(config, dummy_load=False) 
        #To test
        return model



from setuptools import setup, find_packages

setup(
    name='hqq',
    version='0.1.1',    
    description='Half-Quadratic Quantization (HQQ)',
    url='https://github.com/mobiusml/hqq/',
    author='Dr. Hicham Badri',
    author_email='hicham@mobiuslabs.com',
    license='Apache 2',
    packages=['hqq', 'hqq/core', 'hqq/engine', 'hqq/models', 'hqq/models/hf', 'hqq/models/timm', 'hqq/models/vllm'],
    install_requires=['numpy>=1.24.4','tqdm>=4.64.1', 'torch>=2.1.1', 'huggingface_hub', 'accelerate', 'timm', 'transformers>=4.36.1', 'termcolor'], #add vllm/langchain?
)



from ..base import *
from .base  import *

#Patch LLama functions
class LLamaPatch(BasePatch):
	#These tags are used to specify the parameters of each layer type. For example, if you want to give different quantization parameters to different layers
	@classmethod
	def get_linear_tags(cls):
		return ['self_attn.q_proj',
				'self_attn.k_proj',
				'self_attn.v_proj',
				'self_attn.o_proj',
				'mlp.gate_proj'   ,
				'mlp.up_proj'     ,
				'mlp.down_proj'   ]

	@classmethod
	def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
		base_model              = model.model
		model.lm_head           = patch_fct(model.lm_head)
		base_model.embed_tokens = patch_fct(base_model.embed_tokens)
		base_model.norm         = patch_fct(base_model.norm)

		layers = base_model.layers
		for i in tqdm(range(len(base_model.layers)), disable=not verbose):
			layers[i].self_attn.rotary_emb     = patch_fct(layers[i].self_attn.rotary_emb)
			layers[i].mlp.act_fn               = patch_fct(layers[i].mlp.act_fn)
			layers[i].input_layernorm          = patch_fct(layers[i].input_layernorm)
			layers[i].post_attention_layernorm = patch_fct(layers[i].post_attention_layernorm)

	@classmethod
	def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
		base_model = model.model
		layers     = base_model.layers 
		for i in tqdm(range(len(layers)), disable=not verbose):
			layers[i].self_attn.q_proj = patch_fct(layers[i].self_attn.q_proj, patch_params['self_attn.q_proj'])
			layers[i].self_attn.k_proj = patch_fct(layers[i].self_attn.k_proj, patch_params['self_attn.k_proj'])
			layers[i].self_attn.v_proj = patch_fct(layers[i].self_attn.v_proj, patch_params['self_attn.v_proj'])
			layers[i].self_attn.o_proj = patch_fct(layers[i].self_attn.o_proj, patch_params['self_attn.o_proj'])
			layers[i].mlp.gate_proj    = patch_fct(layers[i].mlp.gate_proj,    patch_params['mlp.gate_proj'])
			layers[i].mlp.up_proj      = patch_fct(layers[i].mlp.up_proj,      patch_params['mlp.up_proj'])
			layers[i].mlp.down_proj    = patch_fct(layers[i].mlp.down_proj,    patch_params['mlp.down_proj'])


class LlamaHQQ(LLamaPatch, BaseHQQHFModel):
	#layers to ignore when saving the weights
	@classmethod
	def get_ignore_layers(cls, model):
		return ['', 'model', 'model.layers'] + ['model.layers.' + str(i) for i in range(len(model.model.layers))]

	#Create empty model
	@classmethod
	def create_model(cls, save_dir):
		config = transformers.AutoConfig.from_pretrained(cls.get_config_file(save_dir))
		with init_empty_weights():
			model = transformers.LlamaForCausalLM(config)
		return model
